# This file is part of the Smali Emulator.
#
# Copyright(c) 2016 Simone 'evilsocket' Margaritelli
# evilsocket@gmail.com
# http://www.evilsocket.net
#
# This file may be licensed under the terms of of the
# GNU General Public License Version 3 (the ``GPL'').
#
# Software distributed under the License is distributed
# on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
# express or implied. See the GPL for the specific language
# governing rights and limitations.
#
# You should have received a copy of the GPL along with this
# program. If not, go to http://www.gnu.org/licenses/gpl.html
# or write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
import re

# Base class for all Dalvik opcodes (see
# http://pallergabor.uw.hu/androidblog/dalvik_opcodes.html).


class OpCode(object):
    trace = False

    def __init__(self, expression):
        self.expression = re.compile(expression)

    @staticmethod
    def get_int_value(val):
        ptn = re.compile(r'-?0x\w+?[ts]')
        if ptn.match(val):
            return int(val[:-1], 16)
        elif "0x" in val:
            return int(val, 16)

        return int(val)

    def parse(self, line, vm):
        m = self.expression.search(line)
        if m is None:
            return False

        try:
            self.eval(
                vm, *[x if x is not None else x for x in m.groups()])
        except Exception as e:
            if vm.thrown:
                raise e
            vm.exception(e)

            vm.result = None
            for key in m.groups():
                if key in vm.variables:
                    del vm.variables[key]
                    break

        if OpCode.trace is True:
            # if True:
            print("%03d %s" % (vm.pc, line))
            print('Registers : ', vm.variables)
            print('Result    : ', vm.result)
            print('Return    : ', vm.return_v)

        return True


class op_Const(OpCode):

    def __init__(self):
        OpCode.__init__(self, r'^const(/\d+)?\s+(.+),\s*(.+)')

    @staticmethod
    def eval(vm, _, vx, lit):
        vm[vx] = OpCode.get_int_value(lit)
        vm.result = vm[vx]


class op_ConstWide(OpCode):

    def __init__(self):
        OpCode.__init__(self, r'^const-wide(/\d+)? (.+),\s*(.+)')

    @staticmethod
    def eval(vm, _, vx, lit):
        vm[vx] = OpCode.get_int_value(lit)
        vm.result = vm[vx]


class op_ConstString(OpCode):

    def __init__(self):
        OpCode.__init__(self, r'^const-string(?:/jumbo)? (.+),\s*"(.*)"')

    @staticmethod
    def eval(vm, vx, s):
        vm[vx] = s
        vm.result = vm[vx]


class op_Move(OpCode):

    def __init__(self):
        OpCode.__init__(self, r'^move(-object)?[\/from16]* (.+),\s*(.+)')

    @staticmethod
    def eval(vm, _, vx, vy):
        vm[vx] = vm[vy]


class op_MoveResult(OpCode):

    def __init__(self):
        OpCode.__init__(self, r'^move-result(-object)? (.+)')

    @staticmethod
    def eval(vm, _, dest):
        vm[dest] = vm.result


class op_MoveException(OpCode):

    def __init__(self):
        OpCode.__init__(self, r'^move-exception (.+)')

    @staticmethod
    def eval(vm, vx):
        vm[vx] = vm.exceptions.pop()


class op_IfLe(OpCode):

    def __init__(self):
        OpCode.__init__(self, r'^if-le (.+),\s*(.+),\s*(\:.+)')

    @staticmethod
    def eval(vm, vx, vy, label):
        if vm[vx] <= vm[vy]:
            vm.goto(label)


class op_IfGe(OpCode):

    def __init__(self):
        OpCode.__init__(self, r'^if-ge (.+),\s*(.+),\s*(\:.+)')

    @staticmethod
    def eval(vm, vx, vy, label):
        if vm[vx] >= vm[vy]:
            vm.goto(label)


class op_IfGez(OpCode):

    def __init__(self):
        OpCode.__init__(self, r'^if-gez (.+),\s*(\:.+)')

    @staticmethod
    def eval(vm, vx, label):
        if vm[vx] >= 0:
            vm.goto(label)


class op_IfLtz(OpCode):

    def __init__(self):
        OpCode.__init__(self, r'^if-ltz (.+),\s*(\:.+)')

    @staticmethod
    def eval(vm, vx, label):
        if vm[vx] < 0:
            vm.goto(label)


class op_IfGt(OpCode):

    def __init__(self):
        OpCode.__init__(self, r'^if-gt (.+),\s*(.+),\s*(\:.+)')

    @staticmethod
    def eval(vm, vx, vy, label):
        if vm[vx] > vm[vy]:
            vm.goto(label)


class op_IfGtz(OpCode):

    def __init__(self):
        OpCode.__init__(self, r'^if-gtz (.+),\s*(\:.+)')

    @staticmethod
    def eval(vm, vx, label):
        if vm[vx] > 0:
            vm.goto(label)


class op_IfLez(OpCode):

    def __init__(self):
        OpCode.__init__(self, r'^if-lez (.+),\s*(\:.+)')

    @staticmethod
    def eval(vm, vx, label):
        if vm[vx] <= 0:
            vm.goto(label)


class op_IfEq(OpCode):

    def __init__(self):
        OpCode.__init__(self, r'^if-eq (.+),\s*(.+),\s*(\:.+)')

    @staticmethod
    def eval(vm, vx, vy, label):
        if vm[vx] == vm[vy]:
            vm.goto(label)


class op_IfNe(OpCode):

    def __init__(self):
        OpCode.__init__(self, r'^if-ne (.+),\s*(.+),\s*(\:.+)')

    @staticmethod
    def eval(vm, vx, vy, label):
        if vm[vx] != vm[vy]:
            vm.goto(label)


class op_IfLt(OpCode):

    def __init__(self):
        OpCode.__init__(self, r'^if-lt (.+),\s*(.+),\s*(\:.+)')

    @staticmethod
    def eval(vm, vx, vy, label):
        if vm[vx] < vm[vy]:
            vm.goto(label)


class op_IfEqz(OpCode):

    def __init__(self):
        OpCode.__init__(self, r'^if-eqz (.+),\s*(\:.+)')

    @staticmethod
    def eval(vm, vx, label):
        if vm[vx] == 0:
            vm.goto(label)


class op_IfNez(OpCode):

    def __init__(self):
        OpCode.__init__(self, r'^if-nez (.+),\s*(\:.+)')

    @staticmethod
    def eval(vm, vx, label):
        if vm[vx] != 0:
            vm.goto(label)


class op_ArrayLength(OpCode):

    def __init__(self):
        OpCode.__init__(self, r'array-length (.+),\s*(.+)')

    @staticmethod
    def eval(vm, vx, vy):
        vm[vx] = len(vm[vy])


class op_ArrayFillData(OpCode):

    def __init__(self):
        OpCode.__init__(self, r'fill-array-data (.+),\s*(.+)')

    @staticmethod
    def eval(vm, vx, label):
        vm[vx] = vm.array_data[label]["elements"]


class op_Aget(OpCode):

    def __init__(self):
        OpCode.__init__(self, r'^aget[\-a-z]* (.+),\s*(.+),\s*(.+)')

    @staticmethod
    def eval(vm, vx, vy, vz):
        arr = vm[vy]
        idx = vm[vz]
        vm[vx] = arr[idx]
        vm.result = vm[vx]


class op_AddIntLit(OpCode):

    def __init__(self):
        OpCode.__init__(self, r'^add-int/lit\d+ (.+),\s*(.+),\s*(.+)')

    @staticmethod
    def eval(vm, vx, vy, lit):
        vm[vx] = eval("%s + %s" % (vm[vy], lit))


class op_MulIntLit(OpCode):

    def __init__(self):
        OpCode.__init__(self, r'^mul-int/lit\d+ (.+),\s*(.+),\s*(.+)')

    @staticmethod
    def eval(vm, vx, vy, lit):
        vm[vx] = eval("%s * %s" % (vm[vy], lit))


class op_XorInt2Addr(OpCode):

    def __init__(self):
        OpCode.__init__(self, r'^xor-int(/2addr)? (.+),\s*(.+)')

    @staticmethod
    def eval(vm, _, vx, vy):
        # test if vm[vy] is a char instead of an int
        if isinstance(vm[vy], int):
            vm[vx] ^= int(vm[vy])
        else:
            vm[vx] ^= ord(vm[vy])


class op_DivInt(OpCode):

    def __init__(self):
        OpCode.__init__(self, r'^div-int (.+),\s*(.+),\s*(.+)')

    @staticmethod
    def eval(vm, vx, vy, vz):
        vm[vx] = vm[vy] / vm[vz]


class op_AddInt(OpCode):

    def __init__(self):
        OpCode.__init__(self, r'^add-int (.+),\s*(.+),\s*(.+)')

    @staticmethod
    def eval(vm, vx, vy, vz):
        vm[vx] = vm[vy] + vm[vz]


class op_SubInt(OpCode):

    def __init__(self):
        OpCode.__init__(self, r'^sub-int (.+),\s*(.+),\s*(.+)')

    @staticmethod
    def eval(vm, vx, vy, vz):
        vm[vx] = vm[vy] - vm[vz]


class op_MulInt(OpCode):

    def __init__(self):
        OpCode.__init__(self, r'^mul-int (.+),\s*(.+),\s*(.+)')

    @staticmethod
    def eval(vm, vx, vy, vz):
        vm[vx] = vm[vy] * vm[vz]


class op_RemInt(OpCode):

    def __init__(self):
        OpCode.__init__(self, r'^rem-int (.+),\s*(.+),\s*(.+)')

    @staticmethod
    def eval(vm, vx, vy, vz):
        vm[vx] = vm[vy] % vm[vz]


class op_AndInt(OpCode):

    def __init__(self):
        OpCode.__init__(self, r'^and-int (.+),\s*(.+),\s*(.+)')

    @staticmethod
    def eval(vm, vx, vy, vz):
        vm[vx] = vm[vy] & vm[vz]


class op_OrInt(OpCode):

    def __init__(self):
        OpCode.__init__(self, r'^or-int (.+),\s*(.+),\s*(.+)')

    @staticmethod
    def eval(vm, vx, vy, vz):
        vm[vx] = vm[vy] | vm[vz]


class op_GoTo(OpCode):

    def __init__(self):
        OpCode.__init__(self, r'^goto(/\d+)? (:.+)')

    @staticmethod
    def eval(vm, _, label):
        vm.goto(label)


class op_NewInstance(OpCode):

    def __init__(self):
        OpCode.__init__(self, r'^new-instance (.+),\s*(.+)')

    @staticmethod
    def eval(vm, vx, klass):
        vm[vx] = vm.new_instance(klass)


class op_NewArray(OpCode):

    def __init__(self):
        OpCode.__init__(self, r'^new-array (.+),\s*(.+),\s*(.+)')

    @staticmethod
    def eval(vm, vx, vy, klass):
        vm[vx] = [""] * vm[vy]


class op_APut(OpCode):

    def __init__(self):
        OpCode.__init__(self, r'^aput(-[a-z]+)? (.+),\s*(.+),\s*(.+)')

    @staticmethod
    def eval(vm, _, vx, vy, vz):
        idx = int(vm[vz])
        arr = vm[vy]
        val = vm[vx]
        if len(arr) > idx:
            arr[idx] = val
        elif idx == len(arr):
            arr.append(val)
        vm[vy] = arr


class op_Invoke(OpCode):

    def __init__(self):
        OpCode.__init__(self, r'^invoke-([a-z]+) \{(.*)\},\s*(.+)')

    @staticmethod
    def eval(vm, _, args, call):
        args = list(map(str.strip, args.split(',')))
        this = args[0]
        args = args[1:]
        klass, method = call.split(';->')

        vm.invoke(this, klass, method, args)


class op_IntToType(OpCode):

    def __init__(self):
        OpCode.__init__(self, r'^int-to-([a-z]+) (.+),\s*(.+)')

    @staticmethod
    def eval(vm, ctype, vx, vy):
        if ctype == 'char':
            vm[vx] = chr(vm[vy] & 0xFF)

        else:
            raise RuntimeError("Unsupported type '%s'." % ctype)


class op_SPut(OpCode):

    def __init__(self):
        # sput-object v9, Lcom/whatsapp/messaging/a;->z:[Ljava/lang/String;
        # aput-object v6, v8, v7
        OpCode.__init__(self, r'^sput-object+\s(.+),\s*(.+)')

    @staticmethod
    def eval(vm, vx, staticVariableName):
        vm.variables[staticVariableName] = vm.variables[vx]


class op_SGet(OpCode):

    def __init__(self):
        # sput-object v9, Lcom/whatsapp/messaging/a;->z:[Ljava/lang/String;
        # aput-object v6, v8, v7
        OpCode.__init__(self, r'^sget-object+\s(.+),\s*(.+)')

    @staticmethod
    def eval(vm, vx, staticVariableName):
        vm.variables[vx] = vm.variables[staticVariableName]


class op_Return(OpCode):

    def __init__(self):
        OpCode.__init__(self, r'^return(-[a-z]*)*\s*(.+)*')

    @staticmethod
    def eval(vm, ctype, vx):
        if (ctype is None and vx is None) or ctype == '-void':
            vm.return_v = None
            vm.stop = True
        elif ctype in ('-wide', '-object') or (ctype is None and vx is not None):
            vm.return_v = vm[vx]
            vm.stop = True

        else:
            raise RuntimeError("Unsupported return type.")


class op_RemIntLit(OpCode):

    def __init__(self):
        OpCode.__init__(self, r'^rem-int/lit\d+ (.+),\s*(.+),\s*(.+)')

    @staticmethod
    def eval(vm, vx, vy, lit):
        vm[vx] = int(vm[vy]) % OpCode.get_int_value(lit)


class op_PackedSwitch(OpCode):

    def __init__(self):
        OpCode.__init__(self, r'^packed-switch (.+),\s*(.+)')

    @staticmethod
    def eval(vm, vx, table):
        val = vm[vx]
        switch = vm.packed_switches.get(table, {})
        cases = switch.get('cases', [])
        case_idx = val - switch.get('first_value')

        if case_idx >= len(cases) or case_idx < 0:
            return

        case_label = cases[case_idx]

        vm.goto(case_label)



# -*- coding: utf-8 -*-
# This file is part of the Smali Emulator.
#
# Copyright(c) 2016 Simone 'evilsocket' Margaritelli
# evilsocket@gmail.com
# http://www.evilsocket.net
#
# This file may be licensed under the terms of of the
# GNU General Public License Version 3 (the ``GPL'').
#
# Software distributed under the License is distributed
# on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
# express or implied. See the GPL for the specific language
# governing rights and limitations.
#
# You should have received a copy of the GPL along with this
# program. If not, go to http://www.gnu.org/licenses/gpl.html
# or write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
import sys
import time

from smaliemu.opcodes import *
from smaliemu.vm import VM
from smaliemu.source import Source
from smaliemu.snippet import Snippet
from smaliemu.preprocessors import TryCatchPreprocessor
from smaliemu.preprocessors import PackedSwitchPreprocessor
from smaliemu.preprocessors import ArrayDataPreprocessor
from smaliemu.exception import UnsupportedOpcodeError


# Holds some statistics.
class Stats(object):

    def __init__(self, vm):
        self.opcodes = len(vm.opcodes)
        self.preproc = 0
        self.execution = 0
        self.steps = 0

    def __repr__(self):
        s = "opcode handlers    : %d\n" % self.opcodes
        s += "preprocessing time : %d ms\n" % self.preproc
        s += "execution time     : %d ms\n" % self.execution
        s += "execution steps    : %d\n" % self.steps
        return s


# The main emulator class.
class Emulator(object):

    def __init__(self):
        # Instance of the virtual machine.
        self.vm = None
        # Instance of the source file.
        self.source = None
        self.snippet = None
        # Instance of the statistics object.
        self.stats = None
        # Code preprocessors.
        self.preprocessors = [
            TryCatchPreprocessor, PackedSwitchPreprocessor,
            ArrayDataPreprocessor
        ]
        # Opcodes handlers.
        self.opcodes = []

        # Dynamically load opcode handlers.
        # TODO: Implement missing opcodes.
        for entry in dir(sys.modules['smaliemu.opcodes']):
            if entry.startswith('op_'):
                self.opcodes.append(globals()[entry]())

    def __preprocess(self, smali_code):
        """
        Start the preprocessing phase which will save all the labels and their
        line index for fast lookups while jumping and will pre parse all the
        try/catch directives.
        """
        next_line = None
        smali_code.lines = list(map(str.strip, smali_code.lines))
        for index, line in enumerate(smali_code.lines):
            # we're inside a block which was already processed
            if next_line is not None and index <= next_line:
                next_line = None if index == next_line else next_line
                continue
            elif line == '':  # skip empty lines
                continue
            elif line[0] == ':':  # we've found something to preprocess
                # loop each preprocessors and search for the one responsible
                # to parse this line
                processed = False
                for preproc in self.preprocessors:
                    if preproc.check(line):
                        next_line = preproc.process(self.vm, line, index,
                                                    smali_code.lines)
                        processed = True

                # no preprocessor found, this is a normal label
                if processed is False:
                    self.vm.labels[line] = index

    def __parse_line(self, line):
        # Search for appropriate parser.
        for parser in self.opcodes:
            if parser.parse(line, self.vm):
                return True

        return False

    @staticmethod
    def __should_skip_line(line):
        """
        Helper method used to determine if a line must be skipped/ignored.
        :param line: The line to check.
        :return: True if the line can be ignored, otherwise False.
        """
        return line == "" or line[0] == '#' or line[0] == ':' or line[0] == '.'

    def fatal(self, message):
        """
        Display an error message, the current line being executed and quit.
        :param message: The error message to display.
        """
        print()
        print("-------------------------")
        print("Fatal error on line %03d:\n" % self.vm.pc)
        if self.source:
            print("  %03d %s" % (self.vm.pc, self.source[self.vm.pc - 1]))
        elif self.snippet:
            print("  %03d %s" % (self.vm.pc, self.snippet[self.vm.pc - 1]))

        print("\n%s" % message)

    def get_error_line(self):
        if self.source:
            return "  %03d %s" % (self.vm.pc, self.source[self.vm.pc - 1])
        elif self.snippet:
            return "  %03d %s" % (self.vm.pc, self.snippet[self.vm.pc - 1])

    def run(self, filename, args={}, trace=False):
        """
        Load a smali file and start emulating it.
        :param filename: The path of the file to load and emulate.
        :param args: A dictionary of optional initialization variables for the
                     VM, mostly used for arguments.
        :param trace: If true every opcode being executed will be printed.
        :return: The return value of the emulated method or None if no return-*
                 opcode was executed.
        """
        OpCode.trace = trace
        self.source = Source(filename)
        self.vm = VM(self)
        self.stats = Stats(self)

        self.vm.variables.update(args)

        s = time.time() * 1000
        # Preprocess labels and try/catch blocks for fast lookup.
        self.__preprocess(self.source)
        e = time.time() * 1000
        self.stats.preproc = e - s

        s = time.time() * 1000
        # Loop each line and emulate.
        while self.vm.stop is False and self.source.has_line(self.vm.pc):
            self.stats.steps += 1
            line = self.source[self.vm.pc]
            self.vm.pc += 1

            if self.__should_skip_line(line):
                continue

            elif self.__parse_line(line) is False:
                raise UnsupportedOpcodeError(self.get_error_line())

        e = time.time() * 1000
        self.stats.execution = e - s

        return self.vm.return_v

    def call(self, snippet, trace=False, thrown=True):
        """
        Load a smali file and start emulating it.
        :param snippet: codes snippet in smali format
        :param trace: If true every opcode being executed will be printed.
        :param thrown: If error, emu will raise exception.
        :return: The return value of the emulated method or None if no return-*
                 opcode was executed.
        """
        OpCode.trace = trace
        self.snippet = Snippet(snippet)
        self.vm = VM(self)
        self.stats = Stats(self)
        self.vm.return_v = None
        self.vm.thrown = thrown

        self.vm.variables.copy()

        s = time.time() * 1000
        # Preprocess labels and try/catch blocks for fast lookup.
        self.__preprocess(self.snippet)
        e = time.time() * 1000
        self.stats.preproc = e - s

        s = time.time() * 1000
        # Loop each line and emulate.
        while self.vm.stop is False and self.snippet.has_line(self.vm.pc):
            self.stats.steps += 1
            line = self.snippet[self.vm.pc]
            self.vm.pc += 1

            if self.__should_skip_line(line):
                continue

            elif self.__parse_line(line) is False and thrown:
                raise UnsupportedOpcodeError(self.get_error_line())

        e = time.time() * 1000
        self.stats.execution = e - s

        return self.vm.return_v



# -*- coding: utf-8 -*-
# This file is part of the Smali Emulator.
#
# Copyright(c) 2016 Simone 'evilsocket' Margaritelli
# evilsocket@gmail.com
# http://www.evilsocket.net
#
# This file may be licensed under the terms of of the
# GNU General Public License Version 3 (the ``GPL'').
#
# Software distributed under the License is distributed
# on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
# express or implied. See the GPL for the specific language
# governing rights and limitations.
#
# You should have received a copy of the GPL along with this
# program. If not, go to http://www.gnu.org/licenses/gpl.html
# or write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
from smaliemu.objects.string import String
from smaliemu.objects.string_builder import StringBuilder
from smaliemu.objects.string_buffer import StringBuffer

# This class holds the mapping of Java objects and methods to their Python
# respective.


class ObjectMapping(object):

    def __init__(self):
        self.mapping = {
            String.name(): String.methods(),
            StringBuilder.name(): StringBuilder.methods(),
            StringBuffer.name(): StringBuffer.methods()
        }

    @staticmethod
    def __demangle_class_name(vm, name):
        """
        Demangle a class name.
        :param vm: Instance of the VM.
        :param name: The name of the class to be demangled.
        :return: The demangled class name.
        """
        if name[0] != 'L':
            # vm.emu.fatal("'%s' does not name a class." % name)
            raise RuntimeError("'%s' does not name a class." % name)
        return name[1:].replace('/', '.').replace(';', '')

    def new_instance(self, vm, klass):
        """
        Used by the new-instance opcode.
        :param vm: Instance of the VM.
        :param klass: Mangled class name to instanciate.
        :return: The new class instance.
        """
        class_name = self.__demangle_class_name(vm, klass)

        if class_name in self.mapping:
            if 'new-instance' in self.mapping[class_name]:
                return self.mapping[class_name]['new-instance']()

            else:
                raise RuntimeError(
                    "Unsupported method 'new-instance' for class '%s'." % class_name)
                # vm.emu.fatal("Unsupported method 'new-instance' for class '%s'." % class_name)
        else:
            raise RuntimeError("Unsupported class '%s'." % class_name)
            # vm.emu.fatal("Unsupported class '%s'." % class_name)

    def invoke(self, vm, this, klass, method_name, args):
        """
        Invoke a method ( if mapped ).
        :param vm: Instance of the VM.
        :param this: Identifier of the class instance.
        :param klass: Mangled class name.
        :param method_name: Mangled method name to invoke.
        :param args: Arguments of the method.
        """
        class_name = self.__demangle_class_name(vm, klass)
        if class_name in self.mapping:
            if method_name in self.mapping[class_name]:
                invokeResult = self.mapping[class_name][
                    method_name](vm, this, args)
                if not invokeResult is None:
                    vm.return_v = invokeResult

            else:
                # vm.emu.fatal("Unsupported method '%s' for class '%s'." % ( method_name, class_name ))
                raise RuntimeError(
                    "Unsupported method '%s' for class '%s'." % (method_name, class_name))
        else:
            # vm.emu.fatal("Unsupported class '%s'." % class_name)
            raise RuntimeError("Unsupported class '%s'." % class_name)



class StringBuffer:

    @staticmethod
    def name():
        return 'java.lang.StringBuffer'

    @staticmethod
    def methods():
        return {
            'new-instance': StringBuffer.new_instance,
            '<init>()V': StringBuffer.init,
            '<init>(Ljava/lang/String;)V': StringBuffer.init_from_string,
            'append(Ljava/lang/String;)Ljava/lang/StringBuffer;': StringBuffer.append,
            'append(C)Ljava/lang/StringBuffer;': StringBuffer.append,
            'insert(ILjava/lang/String;)Ljava/lang/StringBuffer;': StringBuffer.insert,
            'toString()Ljava/lang/String;': StringBuffer.tostring
        }

    @staticmethod
    def new_instance():
        return ""

    @staticmethod
    def init(vm, this, args):
        pass

    @staticmethod
    def init_from_string(vm, this, args):
        vm.result = vm[args[0]]
        vm[this] = vm.result

    @staticmethod
    def append(vm, this, args):
        vm[this] += vm[args[0]]
        vm.result = vm[this]

    @staticmethod
    def insert(vm, this, args):
        original = vm[this]
        pos = vm[args[0]]
        new = vm[args[1]]
        vm[this] = original[:pos] + new + original[pos:]
        vm.result = vm[this]

    @staticmethod
    def tostring(vm, this, args):
        vm.result = vm[this]



# This file is part of the Smali Emulator.
#
# Copyright(c) 2016 Simone 'evilsocket' Margaritelli
# evilsocket@gmail.com
# http://www.evilsocket.net
#
# This file may be licensed under the terms of of the
# GNU General Public License Version 3 (the ``GPL'').
#
# Software distributed under the License is distributed
# on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
# express or implied. See the GPL for the specific language
# governing rights and limitations.
#
# You should have received a copy of the GPL along with this
# program. If not, go to http://www.gnu.org/licenses/gpl.html
# or write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
import re
from smaliemu.opcodes import OpCode

# Preprocess try/catch blocks.
class TryCatchPreprocessor:
    @staticmethod
    def check(line):
        return line.startswith( ':try_start_' )

    @staticmethod
    def process(vm, line, index, lines):
        """
        Will search the '.catch' directive and update the VM accordingly.
        """
        # get block identifier
        block_id = line.split('_')[2]
        # search for next pattern:
        #
        #   .catch Ljava/lang/Exception; {:try_start_BLOCK_ID .. :try_end_BLOCK_ID} :LABEL
        #
        expression = '^\.catch [^\s]+ \{:try_start_%s[ \.]+:try_end_%s\}\s*(\:.+)' % (block_id,block_id )
        for nindex, nline in enumerate(lines[index + 1:]):
            # TODO: Save exception type for specific catch.
            m = re.search( expression, nline)
            if m:
                label = m.group(1)
                eindex = nindex + index + 1
                vm.catch_blocks.append((index, eindex, label))
                break

# Preprocess packed-switch blocks.
class PackedSwitchPreprocessor:
    @staticmethod
    def check(line):
        return line.startswith(':pswitch_data')

    @staticmethod
    def process(vm, name, index, lines):
        pswitch   = {"first_value": 0, "cases": []}
        next_line = index

        for nindex, nline in enumerate(lines[index + 1:]):
            if nline.startswith(".packed-switch "):
                pswitch["first_value"] = OpCode.get_int_value(nline.split(' ')[1])

            elif nline.startswith(':pswitch_'):
                pswitch["cases"].append(nline)

            elif nline == '.end packed-switch':
                next_line = index + nindex + 1
                break

            else:
                # vm.fatal("Unexpected line '%s' while preprocessing packed-switch." % nline)
                raise RuntimeError("Unexpected line '%s' while preprocessing packed-switch." % nline)

        vm.packed_switches[name] = pswitch

        # keep preprocessing from the end of this block
        return next_line

# Preprocess array-data blocks.
class ArrayDataPreprocessor:
    @staticmethod
    def check(line):
        return line.startswith(':array_')

    @staticmethod
    def process(vm, name, index, lines):
        array = {"element_width": 0, "elements": []}
        next_line = index

        for nindex, nline in enumerate(lines[index + 1:]):
            if nline.startswith(".array-data "):
                array["element_width"] = OpCode.get_int_value(nline.split(' ')[1])

            elif nline == '.end array-data':
                next_line = index + nindex + 1
                break
            else:
                array["elements"].append(OpCode.get_int_value(nline))

        vm.array_data[name] = array

        # keep preprocessing from the end of this block
        return next_line



# -*- coding: utf-8 -*-
# This file is part of the Smali Emulator.
#
# Copyright(c) 2016 Simone 'evilsocket' Margaritelli
# evilsocket@gmail.com
# http://www.evilsocket.net
#
# This file may be licensed under the terms of of the
# GNU General Public License Version 3 (the ``GPL'').
#
# Software distributed under the License is distributed
# on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
# express or implied. See the GPL for the specific language
# governing rights and limitations.
#
# You should have received a copy of the GPL along with this
# program. If not, go to http://www.gnu.org/licenses/gpl.html
# or write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

# Class to hold the source file data.
class Snippet(object):
    def __init__(self, snippet):
        self.lines = snippet

    def has_line(self, index):
        return True if 0 <= index < len(self.lines) else False

    def __getitem__(self, index):
        return self.lines[index]

    def __setitem__(self, index, line):
        self.lines[index] = line



# -*- coding: utf-8 -*-
# This file is part of the Smali Emulator.
#
# Copyright(c) 2016 Simone 'evilsocket' Margaritelli
# evilsocket@gmail.com
# http://www.evilsocket.net
#
# This file may be licensed under the terms of of the
# GNU General Public License Version 3 (the ``GPL'').
#
# Software distributed under the License is distributed
# on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
# express or implied. See the GPL for the specific language
# governing rights and limitations.
#
# You should have received a copy of the GPL along with this
# program. If not, go to http://www.gnu.org/licenses/gpl.html
# or write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


class String:

    @staticmethod
    def name():
        return 'java.lang.String'

    @staticmethod
    def methods():
        return {
            'new-instance': String.new_instance,
            '<init>([B)V': String.init_from_byte_array,
            '<init>([C)V': String.init_from_char_array,
            'charAt(I)C': String.charat,
            'toCharArray()[C': String.tochararray,
            'intern()Ljava/lang/String;': String.repr_intern,
            'valueOf(Ljava/lang/Object;)Ljava/lang/String;': String.valueof,
            'substring(II)Ljava/lang/String;': String.substring
        }

    @staticmethod
    def repr_intern(vm, this, args):
        return str(vm[this])

    @staticmethod
    def new_instance():
        return ""

    @staticmethod
    def init_from_byte_array(vm, this, args):
        chs = []
        for item in vm[args[0]]:
            chs.append(chr(item))
        vm[this] = "".join(chs)
        vm.result = vm[this]

    @staticmethod
    def init_from_char_array(vm, this, args):
        vm[this] = "".join(vm[args[0]])
        vm.result = vm[this]

    @staticmethod
    def charat(vm, this, args):
        idx = vm[args[0]]
        obj = vm[this]
        vm.result = obj[idx]

    @staticmethod
    def tochararray(vm, this, args):
        vm.result = list(vm[this])

    @staticmethod
    def valueof(vm, this, args):
        vm.result = vm[this]

    @staticmethod
    def substring(vm, this, args):
        start = vm[args[0]]
        end = vm[args[1]]
        vm.result = vm[this][start:end]
        vm[this] = vm.result



# -*- coding: utf-8 -*-
# This file is part of the Smali Emulator.
#
# Copyright(c) 2016 Simone 'evilsocket' Margaritelli
# evilsocket@gmail.com
# http://www.evilsocket.net
#
# This file may be licensed under the terms of of the
# GNU General Public License Version 3 (the ``GPL'').
#
# Software distributed under the License is distributed
# on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
# express or implied. See the GPL for the specific language
# governing rights and limitations.
#
# You should have received a copy of the GPL along with this
# program. If not, go to http://www.gnu.org/licenses/gpl.html
# or write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
from smaliemu.object_mapping import ObjectMapping

# The virtual machine used by the emulator.


class VM(object):

    def __init__(self, emulator):
        # we need the emulator instance in order to call its 'fatal' method.
        self.emu = emulator
        # holds the java->python objects and methods mapping
        self.mapping = ObjectMapping()
        # map of jump labels to opcodes offsets
        self.labels = {}
        # variables container / registers, etc. v1:"str1"
        self.variables = {}
        # try/catch blocks container with opcodes offsets
        self.catch_blocks = []
        # packed switches containers
        self.packed_switches = {}
        # array data blocks
        self.array_data = {}
        # list of thrown exceptions
        self.exceptions = []
        # holds the result of the last method invocation
        self.result = None
        # holds the return value of the method ( used by return-* opcodes )
        self.return_v = None
        # set to true when a return-* opcode is executed
        self.stop = False
        # current opcode index
        self.pc = 0
        # if True, when OpCode eval an unsupported opcode/line, vm will stop.
        self.thrown = False

    def __getitem__(self, name):
        return self.variables[name]

    def __setitem__(self, name, value):
        self.variables[name] = value

    # def fatal(self, message):
    #     self.emu.fatal(message)

    def goto(self, label):
        self.pc = self.labels[label]

    def exception(self, e):
        self.exceptions.append(e)

        # check if this operation is surrounded by a try/catch block
        for block in self.catch_blocks:
            start, end, label = block
            if start <= self.pc <= end:
                self.goto(label)
                return

        # nope, report unhandled exception
        # self.emu.fatal("Unhandled exception '%s'." % str(e))
        # raise RuntimeError("Unhandled exception '%s'." % str(e))

    def new_instance(self, klass):
        return self.mapping.new_instance(self, klass)

    def invoke(self, this, class_name, method_name, args):
        self.mapping.invoke(self, this, class_name, method_name, args)




from setuptools import setup

if __name__ == "__main__":
    setup(
        name="smaliemu",
        version="0.1.6",
        description=("smaliemu"),

        packages=[
            "smaliemu",
            "smaliemu.objects",
        ],

        keywords="",
        license="GPL V3",
        classifiers=[
            "Development Status :: 4 - Beta",
            "License :: OSI Approved :: Apache Software License",
            "Programming Language :: Python :: 3 :: Only",
            "Topic :: Utilities",
        ],

        author="mikusjelly",
        author_email="mikusjelly@gmail.com",
        url="https://github.com/mikusjelly/smaliemu",

    )
